import lmdlang.syntax.tactics_core as tc


__all__ = ('LambdaProgramTactics',)


class LambdaProgramTactics:

    def __init__(self, expression_tactics, *, apply_definitions):
        self._expression = expression_tactics.expression
        self._name = expression_tactics.name
        self._maybe_gap = expression_tactics.maybe_gap
        self._apply_definitions = apply_definitions

    @property
    @tc.tactic
    def definition(self):
        maybe_gap = self._maybe_gap
        assignment_operator = tc.chain(maybe_gap, tc.maybe(':'), '=', maybe_gap)
        end_definition = tc.chain(maybe_gap, ';')

        variable_name, expression = yield tc.chain(
            tc.pick(self._name), assignment_operator, tc.pick(self._expression), end_definition
        )
        return (variable_name, expression)

    @property
    @tc.tactic
    def program(self):
        maybe_gap = self._maybe_gap
        maybe_some_definitions = tc.maybe_some(self.definition, sep=maybe_gap)

        definitions, main_expression = yield tc.chain(
            maybe_gap, tc.pick(maybe_some_definitions), maybe_gap,
            tc.pick(self._expression), maybe_gap
        )
        return self._apply_definitions(definitions, main_expression)



import functools as _functools


__all__ = (
    'tactic', 'apply_tactic', 'TacticFailed',
    'fail', 'catch_fail', 'ResultOrFail',
    'nothing', 'either', 'maybe',
    'some', 'maybe_some',
    'chain', 'pick',
)


# The way tactics are interpreted
# entirely depends on apply_tactic function used
def apply_tactic(tactic, text, *, cursor_index=None):
    if cursor_index is None:
        result, cursor_index = apply_tactic(tactic, text, cursor_index=0)
        if cursor_index < len(text):
            raise TacticFailed(f'parsing failed on position {cursor_index}')
        return result

    if isinstance(tactic, _CompositeTactic):
        result, cursor_index = _apply_composite_tactic(tactic, text, cursor_index=cursor_index)

    elif isinstance(tactic, str):
        result, cursor_index = _apply_string_tactic(tactic, text, cursor_index=cursor_index)

    elif isinstance(tactic, _CatchFailTacticWrapper):
        result, cursor_index = _apply_tactic_catching_fail(tactic.tactic, text, cursor_index=cursor_index)

    elif isinstance(tactic, _TacticFail):
        raise TacticFailed(tactic.fail_info)

    else:
        raise TypeError('not a parsing tactic', tactic)

    return result, cursor_index


def _apply_composite_tactic(tactic, text, *, cursor_index):
    spawned_tactic = tactic()
    latest_tactic_result = None

    while True:
        try:
            next_tactic = spawned_tactic.send(latest_tactic_result)
        except StopIteration as stop_iteration:
            result = stop_iteration.value
            return result, cursor_index
        # May raise TacticFailed
        latest_tactic_result, cursor_index = apply_tactic(next_tactic, text, cursor_index=cursor_index)


def tactic(generator_spawner):
    @_functools.wraps(generator_spawner)
    def make_tactic_by_args(*args, **kwargs):
        ready_generator_spawner = _functools.partial(generator_spawner, *args, **kwargs)
        return _CompositeTactic(ready_generator_spawner)
    return make_tactic_by_args


class _CompositeTactic:

    def __init__(self, generator_spawner):
        self._generator_spawner = generator_spawner

    def __call__(self):
        return self._generator_spawner()


def _apply_string_tactic(tactic, text, *, cursor_index):
    string_required = tactic
    if not text.startswith(string_required, cursor_index):
        raise TacticFailed(f'{repr(string_required)} not found')
    cursor_index += len(string_required)
    return string_required, cursor_index


def _apply_tactic_catching_fail(tactic, text, *, cursor_index):
    try:
        result, cursor_index = apply_tactic(tactic, text, cursor_index=cursor_index)
    except TacticFailed as tactic_fail:
        result = ResultOrFail(tactic_fail.fail_info, failed=True)
    else:
        result = ResultOrFail(result, failed=False)
    return result, cursor_index


class TacticFailed(ValueError):

    def __init__(self, fail_info):
        self._fail_info = fail_info

    @property
    def fail_info(self):
        return self._fail_info


@tactic
def fail(fail_info):
    yield _TacticFail(fail_info)


class _TacticFail:

    def __init__(self, fail_info):
        self._fail_info = fail_info

    @property
    def fail_info(self):
        return self._fail_info


def catch_fail(tactic):
    return _CatchFailTacticWrapper(tactic)


class _CatchFailTacticWrapper:

    def __init__(self, tactic):
        self._tactic = tactic

    @property
    def tactic(self):
        return self._tactic


class ResultOrFail:

    def __init__(self, value, *, failed):
        self._failed = failed
        if self._failed:
            self._fail_info = value
        else:
            self._result = value

    @property
    def failed(self):
        return self._failed

    @property
    def result(self):
        return self._result

    @property
    def fail_info(self):
        return self._fail_info


@tactic
def nothing():
    return
    yield


nothing = nothing()


@tactic
def either(*tactics_to_try):
    for tactic in tactics_to_try:
        applied_tactic = yield catch_fail(tactic)
        if not applied_tactic.failed:
            return applied_tactic.result
    yield fail('All tactics failed')


@tactic
def maybe(tactic, *, otherwise=None):
    parsing = yield catch_fail(tactic)
    if not parsing.failed:
        result = parsing.result
    else:
        result = otherwise
    return result


@tactic
def some(element_tactic, sep=nothing):
    first_element = yield element_tactic

    @tactic
    def read_next():
        yield sep
        parsed_element = yield element_tactic
        return parsed_element

    other_elements = yield _cycle(read_next())

    return [first_element, *other_elements]


@tactic
def _cycle(element_tactic):
    parsed_elements = []
    element_parsing = yield catch_fail(element_tactic)
    while not element_parsing.failed:
        parsed_elements.append(element_parsing.result)
        element_parsing = yield catch_fail(element_tactic)
    return parsed_elements


@tactic
def maybe_some(element, sep=nothing):
    some_elements = some(element, sep=sep)
    result = yield maybe(some_elements, otherwise=[])
    return result


@tactic
def chain(*tactics):
    results = []
    for tactic in tactics:
        if isinstance(tactic, _PickThisTacticResultWrapper):
            results.append((yield tactic.tactic))
        else:
            yield tactic
    return results


def pick(tactic):
    return _PickThisTacticResultWrapper(tactic)


class _PickThisTacticResultWrapper:
    def __init__(self, tactic):
        self._tactic = tactic

    @property
    def tactic(self):
        return self._tactic



#!/usr/bin/env python3

import pathlib as _pathlib
import resource as _resource
import sys as _sys

import lmdlang.language as _language


def main():
    args = _sys.argv[1:]

    if not args:
        code_to_evaluate = _sys.stdin.read()
    elif len(args) == 1:
        lambda_program_path = _pathlib.Path(args[0])
        code_to_evaluate = lambda_program_path.read_text()
    else:
        print('ERROR: Too many arguments passed', file=_sys.stderr)
        return 1

    lang = _language.LambdaLanguage()

    _weaken_recursion_limits()

    term = lang.term_from_code(code_to_evaluate)
    term.normalize()
    output = lang.represent_term(term)

    print(output)

    return 0


def _weaken_recursion_limits():
    _resource.setrlimit(_resource.RLIMIT_STACK, [0x10000000, _resource.RLIM_INFINITY])
    _sys.setrecursionlimit(0x100000)


if __name__ == '__main__':
    _sys.exit(main())



import lmdlang.expression as _expression
import lmdlang.runtime as _runtime
import lmdlang.syntax as _syntax


__all__ = ('LambdaLanguage',)


class LambdaLanguage:

    def __init__(self, *, expr_builder=None, code_reader=None, term_builder=None, expr_representer=None):
        if expr_builder is None:
            expr_builder = self._get_default_expr_builder()
        self._expr_builder = expr_builder

        if code_reader is None:
            code_reader = self._get_default_code_reader()
        self._code_reader = code_reader

        if term_builder is None:
            term_builder = self._get_default_term_builder()
        self._term_builder = term_builder

        if expr_representer is None:
            expr_representer = self._get_default_expr_representer()
        self._expr_representer = expr_representer

    def term_from_code(self, lambda_code):
        main_expression = self._code_reader(lambda_code)
        return main_expression.reinterpret_using(self._term_builder)

    def represent_term(self, term):
        expression = term.as_abstract_expression(expr_builder=self._expr_builder)
        return self._expr_representer(expression)

    def _get_default_expr_builder(self):
        return _expression.basic.LambdaExpressionBuilder()

    def _get_default_code_reader(self):
        expr_tactics = _syntax.classic_expression.LambdaExpressionTactics(self._expr_builder)

        definitions_translator = _expression.definitions.DefinitionsExpressionTranslator(self._expr_builder)

        program_tactics = _syntax.classic_program.LambdaProgramTactics(
            expr_tactics,
            apply_definitions=definitions_translator.apply_definitions
        )

        return lambda code: _syntax.tactics_core.apply_tactic(program_tactics.program, code)

    def _get_default_term_builder(self):
        return _runtime.terms.LambdaTermBuilder()

    def _get_default_expr_representer(self):
        return _syntax.classic_expression.represent_lambda_expression



class VariableNamesManager:

    def __init__(self, *, varnames_pool=None, next_unused_id=None):
        if varnames_pool is None:
            varnames_pool = set()
        self._varnames_pool = varnames_pool

        if next_unused_id is None:
            next_unused_id = 0
        self._next_unused_id = next_unused_id

    def identifier_by_name(self, name):
        self._varnames_pool.add(name)
        return name

    def name_by_identifier(self, identifier):
        return identifier

    def create_new_identifier(self, *, old_identifier=None):
        while str(self._next_unused_id) in self._varnames_pool:
            self._next_unused_id += 1
        result = str(self._next_unused_id)
        self._varnames_pool.add(result)
        return result



class DefinitionsExpressionTranslator:

    def __init__(self, expr_builder):
        self._expr_builder = expr_builder
        self._lam = self._expr_builder.lambda_function
        self._var = self._expr_builder.variable
        self._app = self._expr_builder.application

    def apply_definitions(self, definitions, main_expression):
        if not definitions:
            return main_expression

        definitions_names = [name for name, expr in definitions]
        implementations = [expr for name, expr in definitions]

        objects_tuple = self._create_objects_tuple_by_definitions(definitions_names, implementations)

        main_expression = self._app(
            objects_tuple,
            self._abstract_by_names(main_expression, definitions_names)
        )

        return main_expression

    def _create_objects_tuple_by_definitions(self, definitions_names, implementations):
        lam, var, app = self._lam, self._var, self._app

        def abstract_by_definitions_names(expr):
            return self._abstract_by_names(expr, definitions_names)

        i = 0
        while f'x{i}' in definitions_names:
            i += 1
        objects_tuple_name = f'x{i}'
        i += 1
        while f'x{i}' in definitions_names:
            i += 1
        objects_tuple_callback_name = f'x{i}'

        # Because of possible variables names collision,
        # we cannot directly put any user-defined object inside of a lambda scope.
        # So instead we will create combinator that takes user-defined objects as arguments,
        # and then we will apply it to user-defined objects.
        # Therefore, no variable can be accidentally bound

        objects_tuple_elements = [app(var(objects_tuple_name), var(def_name)) for def_name in definitions_names]
        objects_tuple_by_itself = lam(
            objects_tuple_name,
            self._create_tuple(
                objects_tuple_callback_name,
                objects_tuple_elements
            )
        )

        defined_objects_raw = [abstract_by_definitions_names(impl) for impl in implementations]
        objects_tuple_by_itself = self._pass_arguments(
            abstract_by_definitions_names(objects_tuple_by_itself),
            defined_objects_raw
        )

        objects_tuple = app(self._y_combinator, objects_tuple_by_itself)

        return objects_tuple

    def _pass_arguments(self, applied, arguments):
        result = applied
        for argument in arguments:
            result = self._app(result, argument)
        return result

    def _abstract_by_names(self, expr, names):
        for name in reversed(names):
            expr = self._lam(name, expr)
        return expr

    def _create_tuple(self, callback_name, tuple_values):
        tuple_body = self._pass_arguments(self._var(callback_name), tuple_values)
        return self._lam(callback_name, tuple_body)

    @property
    def _y_combinator(self):
        lam, var, app = self._lam, self._var, self._app
        f = var('f')
        x = var('x')
        y_combinator = lam('f', app(
            lam('x', app(x, x)),
            lam('x', app(f, app(x, x)))
        ))
        return y_combinator



import collections as _collections

import lmdlang.runtime.variable_names_manager as _variable_names_manager


class LambdaTermBuilder:

    def __init__(self, variable_names_manager=None):
        if variable_names_manager is None:
            variable_names_manager = _variable_names_manager.VariableNamesManager()
        self._variable_names_manager = variable_names_manager

    def _term_from_node(self, node):
        return _LambdaTerm(node=node, term_builder=self, variable_names_manager=self._variable_names_manager)

    def lambda_function(self, argument_name, lambda_body):
        argument_name = self._variable_names_manager.identifier_by_name(argument_name)
        return self._term_from_node(_LambdaFunctionNode(argument_name, lambda_body))

    def variable(self, name):
        name = self._variable_names_manager.identifier_by_name(name)
        return self._term_from_node(_VariableNode(name))

    def application(self, applied, argument):
        return self._term_from_node(_ApplicationNode(applied, argument))


_LambdaFunctionNode = _collections.namedtuple('LambdaFunctionNode', 'argument_name body')
_VariableNode = _collections.namedtuple('VariableNode', 'name')
_ApplicationNode = _collections.namedtuple('ApplicationNode', 'applied argument')

_RedirectNode = _collections.namedtuple('RedirectNode', 'target')

_TermNode = (_LambdaFunctionNode, _VariableNode, _ApplicationNode, _RedirectNode)


class _LambdaTerm:

    def __init__(self, *, node, term_builder, variable_names_manager):
        if not isinstance(node, _TermNode):
            raise TypeError('invalid node type', type(node))
        self._term_builder = term_builder
        self._variable_names_manager = variable_names_manager
        self._node = node
        self._is_semi_reduced = False
        self._names_depend_on = self._find_names_depend_on()

    def as_abstract_expression(self, *, expr_builder):
        repr_name = self._variable_names_manager.name_by_identifier

        if self._is_variable_now():
            result = expr_builder.variable(repr_name(self._name))

        elif self._is_lambda_function_now():
            result = expr_builder.lambda_function(
                repr_name(self._argument_name),
                self._body.as_abstract_expression(expr_builder=expr_builder)
            )

        elif self._is_application_now():
            result = expr_builder.application(
                self._applied.as_abstract_expression(expr_builder=expr_builder),
                self._argument.as_abstract_expression(expr_builder=expr_builder)
            )

        else:
            raise RuntimeError

        return result

    def normalize(self):
        terms_to_normalize = targets_stack = [self]
        while targets_stack:
            term = targets_stack.pop()
            term.semi_reduce()
            term, _ = term._drop_outer_lambdas()
            applied, arguments = term._drop_arguments()
            assert applied._is_variable_now()
            targets_stack.extend(arguments)

    def semi_reduce(self):
        terms_to_semi_reduce = targets_stack = [self]
        while targets_stack:
            term = targets_stack[-1]
            if term._is_redirection_now() and not term._without_redirects()._is_semi_reduced:
                targets_stack.append(term._without_redirects())
            elif term._is_lambda_function_now() and not term._body._is_semi_reduced:
                targets_stack.append(term._body)
            elif term._is_application_now() and not term._applied._is_semi_reduced:
                targets_stack.append(term._applied)
            elif term._can_do_outer_reduction():
                term._reduce_by_one_step()
            else:
                term._is_semi_reduced = True
                targets_stack.pop()

    def _collapse_redirects_chain(self):
        if not self._is_redirection_now():
            raise RuntimeError

        latest_redirect = self
        redirects_to_fix = []
        while latest_redirect._node.target._is_redirection_now():
            redirects_to_fix.append(latest_redirect)
            latest_redirect = latest_redirect._node.target

        target = latest_redirect._node.target
        for redirect in redirects_to_fix:
            redirect._node = _RedirectNode(target)

    def _without_redirects(self):
        if not self._is_redirection_now():
            return self
        self._collapse_redirects_chain()
        return self._node.target

    def _node_without_redirects(self):
        return self._without_redirects()._node

    def _is_lambda_function_now(self):
        return isinstance(self._node_without_redirects(), _LambdaFunctionNode)

    def _is_variable_now(self):
        return isinstance(self._node_without_redirects(), _VariableNode)

    def _is_application_now(self):
        return isinstance(self._node_without_redirects(), _ApplicationNode)

    def _is_redirection_now(self):
        return isinstance(self._node, _RedirectNode)

    @property
    def _argument_name(self):
        return self._node_without_redirects().argument_name

    @property
    def _body(self):
        return self._node_without_redirects().body._without_redirects()

    @property
    def _name(self):
        return self._node_without_redirects().name

    @property
    def _applied(self):
        return self._node_without_redirects().applied._without_redirects()

    @property
    def _argument(self):
        return self._node_without_redirects().argument._without_redirects()

    def _find_names_depend_on(self):
        if self._is_variable_now():
            names = frozenset({self._name})
        elif self._is_application_now():
            names = self._applied._names_depend_on | self._argument._names_depend_on
        elif self._is_lambda_function_now():
            names = self._body._names_depend_on
            names -= frozenset({self._argument_name})
        else:
            raise RuntimeError
        return names

    def _may_depend_on_name(self, variable_name):
        return variable_name in self._names_depend_on

    def _can_do_outer_reduction(self):
        return self._is_application_now() and self._applied._is_lambda_function_now()

    def _reduce_by_one_step(self):
        reduced_by_one_step = self._get_reduced_by_one_step()
        if reduced_by_one_step._is_application_now() and not reduced_by_one_step._is_redirection_now():
            self._node = _RedirectNode(reduced_by_one_step)
        else:
            self._node = reduced_by_one_step._node

    def _get_reduced_by_one_step(self):
        assert self._can_do_outer_reduction()

        applied_lambda = self._applied
        target_name = applied_lambda._argument_name
        applied_lambda_body = applied_lambda._body

        if not applied_lambda_body._may_depend_on_name(target_name):
            return applied_lambda_body

        inner_term, bypassed_lambdas_names = applied_lambda_body._drop_outer_lambdas()

        target_name_unused = (
            target_name in bypassed_lambdas_names
            or (inner_term._is_variable_now() and inner_term._name != target_name)
            or not inner_term._may_depend_on_name(target_name)
        )

        if target_name_unused:
            return applied_lambda_body

        target_substitution = self._argument

        new_lambdas_names, lambdas_renamings = self._rename_bypassed_lambdas(
            bypassed_lambdas_names, target_substitution
        )

        inner_term = self._compose_new_inner_term(
            inner_term, target_name, target_substitution, lambdas_renamings
        )

        return inner_term._wrap_in_lambdas(new_lambdas_names)

    def _rename_bypassed_lambdas(self, bypassed_lambdas_names, term_arriving_inside):
        new_lambdas_names = list(bypassed_lambdas_names)
        lambdas_renamings = []
        for lambda_index, lambda_name in enumerate(bypassed_lambdas_names):
            if term_arriving_inside._may_depend_on_name(lambda_name):
                new_lambda_name = self._variable_names_manager.create_new_identifier(old_identifier=lambda_name)
                lambdas_renamings.append((lambda_name, new_lambda_name))
                new_lambdas_names[lambda_index] = new_lambda_name
        return new_lambdas_names, lambdas_renamings

    def _compose_new_inner_term(self, inner_term, target_name, target_substitution, lambdas_renamings):
        if inner_term._is_application_now():
            applied = inner_term._applied
            argument = inner_term._argument
            for old_name, new_name in lambdas_renamings:
                if applied._may_depend_on_name(old_name):
                    applied = applied._rename_free_variable(old_name, new_name)
                if argument._may_depend_on_name(old_name):
                    argument = argument._rename_free_variable(old_name, new_name)
            # Checking applied and argument for _may_depend_on_name(target_name)
            # looks reasonable here, but for unknown reason
            # in practice it slows things down a lot
            applied = applied._substitute(target_name, target_substitution)
            argument = argument._substitute(target_name, target_substitution)
            inner_term = applied._apply(argument)

        elif inner_term._is_variable_now() and inner_term._name == target_name:
            # Nothing to rename, because nobody depends on renamed bypassed lambdas
            inner_term = target_substitution

        return inner_term

    def _drop_outer_lambdas(self):
        term = self
        bypassed_names = []
        while term._is_lambda_function_now():
            bypassed_names.append(term._argument_name)
            term = term._body
        return term, bypassed_names

    def _drop_arguments(self):
        term = self

        arguments = []
        while term._is_application_now():
            arguments.append(term._argument)
            term = term._applied
        arguments.reverse()
        applied = term

        return applied, arguments

    def _wrap_in_lambdas(self, lambdas_names):
        term = self
        for argument_name in reversed(lambdas_names):
            term = term._abstract_by(argument_name)
        return term

    def _substitute(self, name, value):
        return self._abstract_by(name)._apply(value)

    def _rename_free_variable(self, old_name, new_name):
        return self._substitute(old_name, self._term_builder.variable(new_name))

    def _apply(self, argument):
        return self._term_builder.application(self, argument)

    def _abstract_by(self, name):
        return self._term_builder.lambda_function(name, self)



import string

import lmdlang.syntax.tactics_core as tc


__all__ = ('LambdaExpressionTactics',)


class LambdaExpressionTactics:

    def __init__(self, expr_builder):
        self._expr_builder = expr_builder

    _gap_char = tc.either(*string.whitespace)
    maybe_gap = tc.maybe_some(_gap_char)

    _name_char = tc.either(*string.ascii_letters, *string.digits, '_')

    @property
    @tc.tactic
    def name(self):
        chars_sequence = yield tc.some(self._name_char)
        return ''.join(chars_sequence)

    @property
    @tc.tactic
    def _variable(self):
        return self._expr_builder.variable((yield self.name))

    @property
    @tc.tactic
    def _lambda_function(self):
        argument_name, body = yield tc.chain(
            '\\', tc.pick(self.name), '.', self.maybe_gap, tc.pick(self.expression)
        )
        return self._expr_builder.lambda_function(argument_name, body)

    @property
    @tc.tactic
    def _brackets(self):
        content, = yield tc.chain('(', tc.pick(self.expression), ')')
        return content

    @property
    def _application_chain_element(self):
        return tc.either(self._variable, self._lambda_function, self._brackets)

    @property
    @tc.tactic
    def _application_chain(self):
        applied, *arguments = yield tc.some(self._application_chain_element, sep=self.maybe_gap)
        for argument in arguments:
            applied = self._expr_builder.application(applied, argument)
        return applied

    @property
    @tc.tactic
    def expression(self):
        yield self.maybe_gap
        result = yield self._application_chain
        yield self.maybe_gap
        return result


def represent_lambda_expression(expression, *, is_argument=False, is_chain_end=True):
    expr = expression
    repr_expr = represent_lambda_expression

    if expr.is_application():
        applied_repr = repr_expr(expr.applied, is_argument=False, is_chain_end=False)
        argument_repr = repr_expr(expr.argument, is_argument=True, is_chain_end=is_chain_end)
        result = f'{applied_repr} {argument_repr}'
        if is_argument:
            result = f'({result})'

    elif expr.is_lambda_function():
        body_repr = repr_expr(expr.body)
        result = f'\\{expr.argument_name}. {body_repr}'
        if not is_chain_end:
            result = f'({result})'

    elif expr.is_variable():
        result = f'{expr.name}'

    else:
        raise TypeError('unknown lambda expression type')

    return result



import abc as _abc
from abc import abstractmethod as _abstractmethod
import enum as _enum


__all__ = ['ILambdaExpression', 'ILambdaExpressionBuilder']


class ILambdaExpression(_abc.ABC):

    ExprTypes = _enum.Enum('ExprTypes', ['LAMBDA_FUNCTION', 'VARIABLE', 'APPLICATION'])

    @property
    @_abstractmethod
    def expr_type(self) -> ExprTypes:
        pass

    def is_lambda_function(self):
        return self.expr_type is self.ExprTypes.LAMBDA_FUNCTION

    def is_variable(self):
        return self.expr_type is self.ExprTypes.VARIABLE

    def is_application(self):
        return self.expr_type is self.ExprTypes.APPLICATION

    @property
    @_abstractmethod
    def argument_name(self):
        pass

    @property
    @_abstractmethod
    def body(self):
        pass

    @property
    @_abstractmethod
    def name(self):
        pass

    @property
    @_abstractmethod
    def applied(self):
        pass

    @property
    @_abstractmethod
    def argument(self):
        pass

    @_abstractmethod
    def reinterpret_using(self, instructions):
        pass


class ILambdaExpressionBuilder(_abc.ABC):

    @_abstractmethod
    def lambda_function(self, argument_name, body) -> ILambdaExpression:
        pass

    @_abstractmethod
    def variable(self, name) -> ILambdaExpression:
        pass

    @_abstractmethod
    def application(self, applied, argument) -> ILambdaExpression:
        pass



import lmdlang.expression.interface as _interface


__all__ = ['LambdaExpressionBuilder']


class LambdaExpressionBuilder(_interface.ILambdaExpressionBuilder):

    def lambda_function(self, argument_name, body):
        return _LambdaExpression(
            type=_LambdaExpression.ExprTypes.LAMBDA_FUNCTION,
            argument_name=argument_name,
            body=body
        )

    def variable(self, name):
        return _LambdaExpression(
            type=_LambdaExpression.ExprTypes.VARIABLE,
            name=name
        )

    def application(self, applied, argument):
        return _LambdaExpression(
            type=_LambdaExpression.ExprTypes.APPLICATION,
            applied=applied,
            argument=argument
        )


class _LambdaExpression(_interface.ILambdaExpression):

    def __init__(self, type, **kwargs):
        expr_type = type
        del type

        if expr_type is self.ExprTypes.LAMBDA_FUNCTION:
            self._argument_name = kwargs.pop('argument_name')
            self._body = kwargs.pop('body')

        elif expr_type is self.ExprTypes.VARIABLE:
            self._name = kwargs.pop('name')

        elif expr_type is self.ExprTypes.APPLICATION:
            self._applied = kwargs.pop('applied')
            self._argument = kwargs.pop('argument')

        else:
            raise TypeError(f'invalid expression type: {expr_type}')

        self._expr_type = expr_type

        if kwargs:
            raise TypeError(f'unexpected arguments: {kwargs}')

    @property
    def expr_type(self):
        return self._expr_type

    @property
    def argument_name(self):
        return self._argument_name

    @property
    def body(self):
        return self._body

    @property
    def name(self):
        return self._name

    @property
    def applied(self):
        return self._applied

    @property
    def argument(self):
        return self._argument

    def reinterpret_using(self, instructions):
        if self.is_lambda_function():
            argument_name = self.argument_name
            body = self.body.reinterpret_using(instructions)
            result = instructions.lambda_function(argument_name, body)

        elif self.is_variable():
            result = instructions.variable(self.name)

        elif self.is_application():
            applied = self.applied.reinterpret_using(instructions)
            argument = self.argument.reinterpret_using(instructions)
            result = instructions.application(applied, argument)

        else:
            raise TypeError

        return result
